import java.io.IOException;
import java.util.HashMap;

import javafx.animation.Animation;
import javafx.animation.Interpolator;
import javafx.animation.RotateTransition;
import javafx.application.Application;
import javafx.geometry.Point3D;
import javafx.scene.AmbientLight;
import javafx.scene.Group;
import javafx.scene.PerspectiveCamera;
import javafx.scene.PointLight;
import javafx.scene.Scene;
import javafx.scene.input.KeyCode;
import javafx.scene.paint.Color;
import javafx.scene.paint.PhongMaterial;
import javafx.scene.shape.Box;
import javafx.scene.shape.Cylinder;
import javafx.scene.transform.Rotate;
import javafx.stage.Stage;
import javafx.util.Duration;

/*
 * EggHunt.java
 *
 * Use JavaFX to explore a 3D maze.
 * 
 * Find the (literal) Easter eggs.
 *
 * written by mike slattery - apr 2020
 */

public class EggHunt extends Application {
		
    int[][] map;
    public static final double BLOCK_WIDTH = 10.0;
     
    int loc_x,loc_z;
    int dir = 0;
    int[] deltas = {0, -1, 1, 0, 0, 1, -1, 0};
    
    final int NEGGS = 3; 
    int egg_count = 0;
    int score = 0;
    ObjView egg_drvr = null;
    HashMap<String,Group> egg_list = new HashMap<String,Group>();
    
    private PerspectiveCamera camera;
 	private Group cameraDolly;
 	private final double cameraQuantity = 2.0;
 	private final double sceneWidth = 800;
 	private final double sceneHeight = 600;
 	
 	private void constructWorld(Group root) {
		AmbientLight light = new AmbientLight(Color.rgb(100, 100, 100));
		root.getChildren().add(light);
		
		PointLight pl = new PointLight();
		pl.setTranslateX(100);
		pl.setTranslateY(-100);
		pl.setTranslateZ(-100);
		root.getChildren().add(pl);
		
		PointLight pl2 = new PointLight(Color.rgb(0,255,0));
		pl2.setTranslateX(-100);
		pl2.setTranslateY(-100);
		pl2.setTranslateZ(40);
		root.getChildren().add(pl2);
		
		Box floor = new Box((map.length+4)*BLOCK_WIDTH,1,(map[0].length+4)*BLOCK_WIDTH);
		floor.setTranslateX((map.length+2)*BLOCK_WIDTH/2);
		floor.setTranslateY(6);
		floor.setTranslateZ(-(map[0].length+2)*BLOCK_WIDTH/2);
		final PhongMaterial floorMaterial = new PhongMaterial();
		floorMaterial.setDiffuseColor(Color.BROWN);
		floorMaterial.setSpecularColor(Color.WHITE);
		floor.setMaterial(floorMaterial);
		root.getChildren().add(floor);

		Color color = Color.rgb(255, 200, 80);
		final PhongMaterial blockMaterial = new PhongMaterial();
		blockMaterial.setDiffuseColor(color);
		blockMaterial.setSpecularColor(Color.WHITE);
		for (int j = 0; j < map[0].length; j++)
		  for (int i = 0; i < map.length; i++)
			if (map[i][j] != 0)
			{
				Box block = new Box(BLOCK_WIDTH, BLOCK_WIDTH, BLOCK_WIDTH);
				block.setMaterial(blockMaterial);
				block.setTranslateX(BLOCK_WIDTH+i*BLOCK_WIDTH);
				block.setTranslateZ(-(BLOCK_WIDTH+j*BLOCK_WIDTH));
				root.getChildren().add(block);
			}

    } // end of constructWorld method of Labyrinth


    @Override
	public void start(Stage primaryStage) {
    	// Generate maze
    	MazeGenerator mg = new MazeGenerator();
		map = mg.genMaze(14, 12);
		for (int j = 0; j < map[0].length; j++)
		{
		  for (int i = 0; i < map.length; i++)
		    System.out.print(map[i][j]);
		  System.out.println();
		}
		
		// Build your Scene and Camera
		Group sceneRoot = new Group();
		constructWorld(sceneRoot);
		
		// Add random eggs
		for (int i = 0; i < NEGGS; i++) {
			Group egg = placeRandomEgg();
			sceneRoot.getChildren().add(egg);
		}
				
        // This will move the Camera to a random
        // location in the maze.
        // Find a spot without a block:
        while(true)
        {
          loc_x = (int)(Math.random()*14);
          loc_z = (int)(Math.random()*12);
          if (map[loc_x][loc_z] == 0)
            break;
        }
        System.out.println("("+loc_x+","+loc_z+")");
        double vp_x = BLOCK_WIDTH+loc_x*BLOCK_WIDTH;
        double vp_z = -(BLOCK_WIDTH+loc_z*BLOCK_WIDTH);

		Scene scene = new Scene(sceneRoot, sceneWidth, sceneHeight, true);
		scene.setFill(Color.SKYBLUE);
		camera = new PerspectiveCamera(true);
		camera.setNearClip(0.1);
		camera.setFarClip(10000.0);
		scene.setCamera(camera);
		Group avatar = new Group();
		camera.setTranslateZ(-1.4*BLOCK_WIDTH);
		PhongMaterial cMaterial = new PhongMaterial();
		cMaterial.setDiffuseColor(new Color(0,0,1,0.4));
		Cylinder c = new Cylinder(1,3);
		c.setMaterial(cMaterial);
		c.setTranslateY(3);
		avatar.getChildren().add(camera);
		avatar.getChildren().add(c);
		// translations through dolly
		cameraDolly = new Group();
		cameraDolly.setTranslateZ(vp_z);
		cameraDolly.setTranslateX(vp_x);
		cameraDolly.getChildren().add(avatar);
		sceneRoot.getChildren().add(cameraDolly);
		// rotation transform
		Rotate yRotate = new Rotate(0, 0, 0, 0, Rotate.Y_AXIS);
		avatar.getTransforms().addAll(yRotate);
		
		
		// Use keyboard to control camera position
		scene.setOnKeyPressed(event -> {
			double change = cameraQuantity;
			// What key did the user press?
			KeyCode keycode = event.getCode();

			// Check for movement
			if (keycode == KeyCode.UP || keycode == KeyCode.DOWN) {
				int new_x, new_z;
				
				if (keycode == KeyCode.UP) {
					new_x = loc_x + deltas[2*dir];
					new_z = loc_z + deltas[2*dir+1];		
				} else {
					new_x = loc_x - deltas[2*dir];
					new_z = loc_z - deltas[2*dir+1];
				}
				System.out.println("("+loc_x+","+loc_z+")->("+new_x+","+new_z+")");
				int mval = map[new_x][new_z];
				if (mval == 0 || mval == 3) {
					loc_x = new_x; loc_z = new_z;
					cameraDolly.setTranslateZ(-(BLOCK_WIDTH+loc_z*BLOCK_WIDTH));
					cameraDolly.setTranslateX(BLOCK_WIDTH+loc_x*BLOCK_WIDTH);
				}
				if (mval == 3) {
					score++;
					System.out.println("*** SCORE = "+score+" ***");
					map[loc_x][loc_z] = 0;
					egg_list.get(""+loc_x+","+loc_z).setVisible(false);
					if (score == egg_count)
						System.out.println("***** YOU WIN!!! *****");
				}
					//System.out.println("blocked");
			}
			// Check for turning
			if (keycode == KeyCode.LEFT){
				dir = (dir + 3) % 4;
				yRotate.setAngle(((yRotate.getAngle() - 90) % 360 + 540) % 360 - 180); // -
			}
			if (keycode == KeyCode.RIGHT){
				dir = (dir + 1) % 4;
				yRotate.setAngle(((yRotate.getAngle() + 90) % 360 + 540) % 360 - 180); // +
			}
			//For debug
//			Point3D delta = null;
//			
//			if (keycode == KeyCode.A) {
//				delta = new Point3D(-change, 0, 0);
//			}
//			if (keycode == KeyCode.D) {
//				delta = new Point3D(change, 0, 0);
//			}
//			if (keycode == KeyCode.W) {
//				delta = new Point3D(0, -change, 0);
//			}
//			if (keycode == KeyCode.S) {
//				delta = new Point3D(0, change, 0);
//			}
//			if (delta != null) {
//				Point3D delta2 = avatar.localToParent(delta);
//				cameraDolly.setTranslateX(cameraDolly.getTranslateX() + delta2.getX());
//				cameraDolly.setTranslateY(cameraDolly.getTranslateY() + delta2.getY());
//				cameraDolly.setTranslateZ(cameraDolly.getTranslateZ() + delta2.getZ());
//			}		
		});

		primaryStage.setTitle("Egg Hunt");
		primaryStage.setScene(scene);
		primaryStage.show();
	}
    
    Group placeRandomEgg(){
    	int egg_x, egg_z;
    	Group egg = null;
		
    	// Find an empty location
		while(true)
        {
          egg_x = (int)(Math.random()*14);
          egg_z = (int)(Math.random()*12);
          if (map[egg_x][egg_z] == 0)
            break;
        }
		map[egg_x][egg_z] = 3;
		
        System.out.println("Egg:("+egg_x+","+egg_z+")");
        double e_x = BLOCK_WIDTH+egg_x*BLOCK_WIDTH;
        double e_z = -(BLOCK_WIDTH+egg_z*BLOCK_WIDTH);
        if (egg_drvr == null){
			egg_drvr = new ObjView();
			try {
				egg_drvr.load(ClassLoader.getSystemResource("egg.obj").toString());
			} catch (IOException e) {
				System.out.println("Trouble loading model");
				e.printStackTrace();
			}
			egg = egg_drvr.getRoot();
        } else {
        	ObjView drvr2 = new ObjView(egg_drvr);
    		egg = drvr2.getRoot();
        }
        // Position egg
		egg.setScaleX(2);
		egg.setScaleY(-2);
		egg.setScaleZ(-2);
		egg.setTranslateX(e_x);
		egg.setTranslateY(-4);
		egg.setTranslateZ(e_z);

		RotateTransition rt = new RotateTransition(Duration.millis(2000), egg);
		rt.setByAngle(100);
		rt.setAxis(Rotate.Y_AXIS);
		rt.setInterpolator(Interpolator.LINEAR);
		rt.setCycleCount(Animation.INDEFINITE);
		rt.setAutoReverse(true);
		rt.play();
		// record egg
		egg_count++;
		egg_list.put(""+egg_x+","+egg_z, egg);
		return egg;
    }

	public static void main(String[] args) {
		launch(args);
	}

} // end of class EggHunt
