/*
 * MazeGenerator.java
 *
 * Use depth-first search to generate a random rectangular
 * maze.
 *
 * written by mike slattery - mar 1999
 * modified to update Thread handling, mcs - mar 2002
 * modified for block maze, mcs - nov 2003
 */
import java.util.*;
import java.awt.Point;

public class MazeGenerator
{
	// Use simpler constants for block directions
	// (than original Dfs)
	public static final int WTOP = 0;
	public static final int WRGT = 1;
	public static final int WBOT = 2;
	public static final int WLFT = 3;

	int cells[][];
	Point current_cell;
	Stack<Point> hold;

	public int[][] genMaze(int gridw, int gridh)
	{
	  // Return a gridw by gridh maze (with 0 for
	  // open space, non-zero for blocks).
	  // The gridw is considered the limit for the
	  // first coordinate.
		int x, y;
		Point nbr;
		
		cells = new int[gridw][gridh];
		
		// Put blocks everywhere
		for (x = 0; x < gridw; x++)
			for (y = 0; y < gridh; y++)
				cells[x][y] = 1;
		/* Then, mark the borders
		 */
		for (y = 0; y < gridh; y++)
		{
			cells[0][y] = 2;
			cells[gridw-1][y] = 2;
		}
		for (x = 0; x < gridw; x++)
		{
			cells[x][0] = 2;
			cells[x][gridh-1] = 2;
		}

		/*
		 * Implement depth-first search to build maze
		 */
		current_cell = new Point(rnd(gridw-2),rnd(gridh-2));
		removeBlock(current_cell);
		hold = new Stack<Point>();
		while (true)
		{
			nbr = findNewNbr(current_cell);
			if (nbr == null)
			{
				if (hold.isEmpty())
					break;
				current_cell = (Point)(hold.pop());
			}
			else
			{
				removeBlock(nbr);
				hold.push(current_cell);
				current_cell = nbr;
			}
		}
		return cells;
	}
	
	Point findNewNbr(Point p)
	{
		/* Return a neighboring block which is
		 * removable (or null if none such).
		 */
		int d = rnd(4)-1;
		int k = 0;
		while (k < 4)
		{
			Point nbr = getNbr(p, d);
			if (isRemovable(nbr))
			  return nbr;
			d = (d+1) % 4;
			k++;
		}
		return null;
	}
	
	/*
	 * The following routines provide access to the underlying
	 * maze data structure.
	 */	
	boolean isRemovable(Point p)
	{
	  // Return whether or not the block at p
	  // can be removed without breaking through
	  // to the existing maze.
	  Point nbr;
	  
	  if (cells[p.x][p.y] != 1)
	    return false;
	  int sum = 0;
	  for (int d = 0; d < 4; d++)
	  {
	     nbr = getNbr(p, d);
	     if (cells[nbr.x][nbr.y] != 0)
	       sum++;
	  }
	  return sum == 3;
	}

	void removeBlock(Point p)
	{
		cells[p.x][p.y] = 0;
	}

	Point getNbr(Point p, int d)
	{
	   // Return the neighboring point from p in
	   // the direction d.
		switch(d)
		{
			case WTOP: return new Point(p.x, p.y-1);
			case WRGT: return new Point(p.x+1, p.y);
			case WBOT: return new Point(p.x, p.y+1);
			case WLFT: return new Point(p.x-1, p.y);
		}
		return null;
	}

	int rnd(int n)
	{
	  // Return a random integer from 1 to n
		return (int)(Math.random()*n+1);
	}
}
